__all__ = ()


import reprlib, warnings
from datetime import datetime

from scarletio import Compound, Theory

from ...channel import Channel
from ...color import Color
from ...http import DiscordHTTPClient, VALID_ICON_MEDIA_TYPES, VALID_ICON_MEDIA_TYPES_EXTENDED
from ...oauth2 import Connection
from ...user import PremiumType
from ...utils import datetime_to_timestamp, get_image_media_type, image_to_base64

from ..request_helpers import get_guild_and_id, get_guild_id, get_channel_guild_id_and_id, validate_timeout_duration



def _assert__guild_profile_edit__nick(nick):
    """
    Asserts the `nick` parameter of ``Client.guild_profile_edit`` method.
    
    Parameters
    ----------
    nick : `Ellipsis`, `None`, `int`
        The new nick of the user.
    
    Raises
    ------
    AssertionError
        - If `nick` was not given neither as `None`, `str`.
        - If `nick` length is out of the expected range [0:32].
    """
    if nick is ...:
        pass
    
    elif nick is None:
        pass
    
    elif isinstance(nick, str):
        nick_length = len(nick)
        if nick_length > 32:
            raise AssertionError(
                f'`nick` length can be in range [0:32], got {nick_length}; {nick!r}.'
            )
    
    else:
        raise AssertionError(
            f'`nick` can be `None`, `str`, got {nick.__class__.__name__}; {nick!r}.'
        )
    
    return True


def _assert__guild_profile_edit__avatar(avatar, client):
    """
    Asserts the `avatar` parameter of ``Client.guild_profile_edit`` method.
    
    Parameters
    ----------
    avatar : `Ellipsis`, `None`, `bytes-like`
        The client's new guild specific avatar.
    client : ``Client``
        The client to query additional client-specific limitations.
    
    Raises
    ------
    AssertionError
        - If `avatar`'s type is incorrect.
        - If `avatar`'s format is incorrect.
    """
    if (avatar is not ...) and (avatar is not None):
        if not AssertionError(avatar, (bytes, bytearray, memoryview)):
            raise TypeError(
                f'`avatar` can be `None`, `bytes-like`, got {avatar.__class__.__name__}; '
                f'{reprlib.repr(avatar)}.'
            )
        
        media_type = get_image_media_type(avatar)
        
        if client.premium_type.value:
            valid_icon_media_types = VALID_ICON_MEDIA_TYPES_EXTENDED
        else:
            valid_icon_media_types = VALID_ICON_MEDIA_TYPES
        
        if media_type not in valid_icon_media_types:
            raise AssertionError(
                f'Invalid `avatar` type for the client: {media_type}, got {reprlib.repr(avatar)}.'
            )
    
    return True


class ClientCompoundClientEndpoints(Compound):
    
    guild_profiles : dict
    http : DiscordHTTPClient
    is_bot : bool
    premium_type : PremiumType
    
    @Theory
    def _set_attributes(self, data): ...
    
    
    async def client_edit(self, **keyword_parameters):
        """
        ``Client.client_edit`` is deprecated and will be removed in 2022 December.
        Please use ``.edit`` instead.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.client_edit` is deprecated and will be removed in '
                f'2022 December. Please use `.edit` instead.'
            ),
            FutureWarning
        )
        return await self.edit(**keyword_parameters)
    
    
    async def edit(
        self, *, avatar=..., banner=..., banner_color=..., bio=..., name=..., # Generic
        password=..., new_password=..., email=..., # User account only
    ):
        """
        Edits the client. Only the provided parameters will be changed. Every parameter what refers to a user
        account is not tested.
        
        This method is a coroutine.
        
        Parameters
        ----------
        avatar : `None`, `bytes-like`, Optional (Keyword only)
            An `'jpg'`, `'png'`, `'webp'` image's raw data. If the client is premium account, then it can be
            `'gif'` as well. By passing `None` you can remove the client's current avatar.
        
        banner : `None`, `bytes-like`, Optional (Keyword only)
            An `'jpg'`, `'png'`, `'webp'`, 'gif'` image's raw data. By passing `None` you can remove the client's
            current avatar.
        
        banner_color : `None`, ``Color``, `int`, Optional (Keyword only)
            The new banner color of the client. By passing it as `None` you can remove the client's current one.
        
        bio : `None`, `str`, Optional (Keyword only)
            The new bio of the client. By passing it as `None`, you can remove the client's current one.
        
        name : `str`, Optional (Keyword only)
            The client's new name.
        
        password : `str`, Optional (Keyword only)
            The actual password of the client.
        
        new_password : `str`, Optional (Keyword only)
            The client's new password.
        
        email : `str`, Optional (Keyword only)
            The client's new email.
        
        Raises
        ------
        TypeError
            - If `avatar` was not given as `None`, neither as `bytes-like`.
            - If `banner` was not given as `None`, neither as `bytes-like`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `name` was given but not as `str`.
            - If `name`'s length is out of range [2:32].
            - If `avatar`'s type in unsettable for the client.
            - If `password` was not given meanwhile the client is not bot.
            - If `password` was not given as `str`.
            - If `email` was given, but not as `str`.
            - If `new_password` was given, but not as `str`.
            - If `bio` is neither `None` nor `str`.
            - If `bio`'s length is out of range [0:190].
            - if `banner_color` is neither `None` nor `int`.
        
        Notes
        -----
        The method's endpoint has long rate limit reset, so consider using timeout and checking rate limits with
        ``RateLimitProxy``.
        
        The `password`, `new_password` and the `email` parameters are only for user accounts.
        """
        assert _assert__guild_profile_edit__avatar(avatar, self)
        
        data = {}
        
        
        if (avatar is not ...):
            if avatar is None:
                avatar_data = None
            else:
                avatar_data = image_to_base64(avatar)
            
            data['avatar'] = avatar_data
        
        
        if (banner is not ...):
            if banner is None:
                banner_data = None
            else:
                if not isinstance(banner, (bytes, bytearray, memoryview)):
                    raise TypeError(
                        f'`banner` can be `None`, `bytes-like`, got {banner.__class__.__name__}; '
                        f'{reprlib.repr(banner)}.'
                    )
                
                if __debug__:
                    media_type = get_image_media_type(banner)
                    
                    if media_type not in VALID_ICON_MEDIA_TYPES_EXTENDED:
                        raise AssertionError(
                            f'Invalid `banner` type for the client: {media_type}; got {reprlib.repr(banner)}.'
                        )
                
                banner_data = image_to_base64(banner)
            
            data['banner'] = banner_data
        
        
        if (banner_color is not ...):
            if __debug__:
                if (banner_color is not None) and (not isinstance(banner_color, int)):
                    raise AssertionError(
                        f'`banner_color` can be `None`, `{Color.__name__}`, `int`, got '
                        f'{banner_color.__name__}; {banner_color!r}.'
                    )
            
            data['accent_color'] = banner_color
        
        
        if (bio is not ...):
            if bio is None:
                bio = ''
            else:
                if __debug__:
                    if not isinstance(bio, str):
                        raise AssertionError(
                            f'`bio` can be `None`, `str`, got {bio.__class__.__name__}; {bio!r}.'
                        )
                    
                    bio_length = len(bio)
                    if bio_length > 190:
                        raise AssertionError(
                            f'`bio` length can be in range [0:190], got {bio_length!r}; {bio!r}.'
                        )
            
            data['bio'] = bio
        
        
        if (name is not ...):
            if __debug__:
                if not isinstance(name, str):
                    raise AssertionError(
                        f'`name` can be `str`, got {name.__class__.__name__}; {name!r}.'
                    )
                
                name_length = len(name)
                if name_length < 2 or name_length > 32:
                    raise AssertionError(
                        f'The length of the name can be in range [2:32], got {name_length}; {name!r}.'
                    )
            
            data['username'] = name
        
        
        if not self.is_bot:
            if __debug__:
                if password is ...:
                    raise AssertionError(
                        f'`password` is must for non bots, got {password!r}.'
                    )
                
                if not isinstance(password, str):
                    raise AssertionError(
                        f'`password` can be `str`, got {password.__class__.__name__}; {password!r}.'
                    )
            
            data['password'] = password
            
            
            if (email is not ...):
                if __debug__:
                    if not isinstance(email, str):
                        raise AssertionError(
                            f'`email` can be `str`, got {email.__class__.__name__}; {email!r}.'
                        )
                
                data['email'] = email
            
            
            if (new_password is not ...):
                if __debug__:
                    if not isinstance(new_password, str):
                        raise AssertionError(
                            f'`new_password` can be `str`, got {new_password.__class__.__name__}; '
                            f'{new_password!r}.'
                        )
                
                data['new_password'] = new_password
        
        
        data = await self.http.client_edit(data)
        self._set_attributes(data)
        
        
        if not self.is_bot:
            self.email = data['email']
            try:
                token = data['token']
            except KeyError:
                pass
            else:
                self.token = token
    
    
    async def client_guild_profile_edit(self, *positional_parameters, **keyword_parameters):
        """
        ``Client.client_guild_profile_edit`` is deprecated and will be removed in 2022 December.
        Please use ``.guild_profile_edit`` instead.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.client_guild_profile_edit` is deprecated and will be removed in '
                f'2022 December. Please use `.guild_profile_edit` instead.'
            ),
            FutureWarning
        )
        return await self.guild_profile_edit(*positional_parameters, **keyword_parameters)
    
    
    async def guild_profile_edit(
        self, guild, *, nick=..., avatar=..., timeout_duration = ..., timed_out_until=..., reason=None
    ):
        """
        Edits the client guild profile in the given guild. Nick and guild specific avatars can be edited on this way.
        
        An extra `reason` is accepted as well, which will show up at the respective guild's audit logs.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : `None`, `int`, ``Guild``
            The guild where the client's nickname will be changed. If `guild` is given as `None`, then the function
            returns instantly.
        
        nick : `None`, `str`, Optional (Keyword only)
            The client's new nickname. Pass it as `None` to remove it. Empty strings are interpreted as `None`.
        
        avatar : `None`, `bytes-like`, Optional (Keyword only)
            The client's new guild specific avatar.
            
            Can be a `'jpg'`, `'png'`, `'webp'` image's raw data. If the client is premium account, then it can be
            `'gif'` as well. By passing `None` you can remove the client's current avatar.
        
        timeout_duration: `None`, `int`, `float`, `timedelta`, `datetime`, Optional (Keyword only)
            The timeout duration of the user in seconds.
            
            Pass it as `None` or as a non-positive duration to remove it.
            
            The max allowed value equals to 28 days.
        
        timed_out_until : `None`, `datetime`, Optional (Keyword only)
            Till when the client is timed out. Pass it as `None` to remove it.
            
            > Deprecated and will be removed in 2022 december. Please use `timeout_duration` instead.
        
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Will show up at the respective guild's audit logs.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild, guild_id = get_guild_and_id(guild)
        
        assert _assert__guild_profile_edit__nick(nick)
        assert _assert__guild_profile_edit__avatar(avatar, self)
        
        data = {}
        
        if (nick is not ...):
            # Non debug mode: Translate empty nick to `None`
            if (nick is not None) and (not nick):
                nick = None
            
            # Check whether we should edit the nick.
            if guild is None:
                # `guild` can be `None` if `guild` parameter was given as `int`.
                should_edit_nick = True
            else:
                try:
                    guild_profile = self.guild_profiles[guild.id]
                except KeyError:
                    # we aren't at the guild probably ->  will raise the request for us, if really
                    should_edit_nick = True
                else:
                    should_edit_nick = (guild_profile.nick != nick)
        
        
            if should_edit_nick:
                data['nick'] = nick
        
        
        if (avatar is not ...):
            if avatar is None:
                avatar_data = None
            else:
                avatar_data = image_to_base64(avatar)
            
            data['avatar'] = avatar_data
        
        
        if (timed_out_until is not ...):
            warnings.warn(
                (
                    f'`timed_out_until` parameter of `{self.__class__.__name__}.guild_profile_edit` is deprecated'
                    f'and will be removed in 2022 December. Please use `timeout_duration` instead.'
                ),
                FutureWarning,
            )
            timeout_duration = timed_out_until
        
        
        if (timeout_duration is not ...):
            timeout_ends_at = validate_timeout_duration(timeout_duration)
            
            if (timeout_ends_at is None):
                timed_out_until_raw = None
            else:
                timed_out_until_raw = datetime_to_timestamp(timeout_ends_at)
            data['communication_disabled_until'] = timed_out_until_raw
        
        
        if data:
            await self.http.client_guild_profile_edit(guild_id, data, reason)
    
    
    async def client_connection_get_all(self):
        """
        ``Client.client_connection_get_all`` is deprecated and will be removed in 2022 December.
        Please use ``.connection_get_all`` instead.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.client_connection_get_all` is deprecated and will be removed in '
                f'2022 December. Please use `.connection_get_all` instead.'
            ),
            FutureWarning
        )
        return await self.connection_get_all()
    
    
    async def connection_get_all(self):
        """
        Requests the client's connections.
        
        This method is a coroutine.
        
        Returns
        -------
        connections : `list` of ``Connection`` objects
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        For a bot account this request will always return an empty list.
        """
        data = await self.http.client_connection_get_all()
        return [Connection(connection_data) for connection_data in data]
    
    
    async def guild_leave(self, guild):
        """
        The client leaves the given guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild from where the client will leave.
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        
        await self.http.guild_leave(guild_id)


    async def join_speakers(self, channel, *, request=False):
        """
        Request to speak or joins the client as a speaker inside of a stage channel. The client must be in the stage
        channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``
            The stage channel to join.
        request : `bool` = `False`, Optional (Keyword only)
            Whether the client should only request to speak.
        
        Raises
        ------
        TypeError
            If `channel` was not given neither as ``Channel`` nor as `tuple` (`int`, `int`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id, channel_id = get_channel_guild_id_and_id(channel, Channel.is_guild_stage)
        
        if request:
            timestamp = datetime_to_timestamp(datetime.utcnow())
        else:
            timestamp = None
        
        data = {
            'suppress': False,
            'request_to_speak_timestamp': timestamp,
            'channel_id': channel_id
        }
        
        await self.http.voice_state_client_edit(guild_id, data)
    
    
    async def join_audience(self, channel):
        """
        Moves the client to the audience inside of the stage channel. The client must be in the stage channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `tuple` (`int`, `int`)
            The stage channel to join.
        
        Raises
        ------
        RuntimeError
            If `channel` is partial.
        TypeError
            If `channel` was not given neither as ``Channel`` nor as `tuple` (`int`, `int`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id, channel_id = get_channel_guild_id_and_id(channel, Channel.is_guild_stage)
        
        data = {
            'suppress': True,
            'channel_id': channel_id
        }
        
        await self.http.voice_state_client_edit(guild_id, data)



__all__ = ('ChannelMetadataGuildMainBase',)

from scarletio import copy_docs, include

from ...core import GUILDS
from ...permission import Permission
from ...permission.permission import (
    PERMISSION_ALL, PERMISSION_MASK_ADMINISTRATOR, PERMISSION_MASK_VIEW_CHANNEL, PERMISSION_NONE
)
from ...permission.utils import PERMISSION_ALLOW_KEY, PERMISSION_DENY_KEY
from ...user import ClientUserBase

from .guild_base import ChannelMetadataGuildBase


parse_permission_overwrites = include('parse_permission_overwrites')


class ChannelMetadataGuildMainBase(ChannelMetadataGuildBase):
    """
    Base class for main guild metadata channels not including thread channels.
    
    Attributes
    ----------
    _permission_cache : `None`, `dict` of (`int`, ``Permission``) items
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    parent_id : `int`
        The channel's parent's identifier.
    name : `str`
        The channel's name.
    permission_overwrites : `dict` of (`int`, ``PermissionOverwrite``) items
        The channel's permission overwrites.
    position : `int`
        The channel's position.
    
    Class Attributes
    ----------------
    type : `int` = `-1`
        The channel's type.
    order_group: `int` = `0`
        The channel's order group used when sorting channels.
    """
    __slots__ = ('permission_overwrites', 'position', )
    
    @copy_docs(ChannelMetadataGuildBase._created)
    def _created(self, channel_entity, client):
        try:
            guild = GUILDS[channel_entity.guild_id]
        except KeyError:
            pass
        else:
            guild.channels[channel_entity.id] = channel_entity
    
    
    @copy_docs(ChannelMetadataGuildBase._delete)
    def _delete(self, channel_entity, client):
        ChannelMetadataGuildBase._delete(self, channel_entity, client)
        
        try:
            guild = GUILDS[channel_entity.guild_id]
        except KeyError:
            pass
        else:
            try:
                del guild.channels[channel_entity.id]
            except KeyError:
                pass
    
    
    @copy_docs(ChannelMetadataGuildBase._compare_attributes_to)
    def _compare_attributes_to(self, other):
        if not ChannelMetadataGuildBase._compare_attributes_to(self, other):
            return False
        
        if self.permission_overwrites != other.permission_overwrites:
            return False
        
        if self.position != other.position:
            return False
        
        return True
    
    
    @copy_docs(ChannelMetadataGuildBase._update_attributes)
    def _update_attributes(self, data):
        ChannelMetadataGuildBase._update_attributes(self, data)
        
        self.permission_overwrites = parse_permission_overwrites(data)
        
        self.position = data.get('position', 0)
    
    
    @copy_docs(ChannelMetadataGuildBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = ChannelMetadataGuildBase._difference_update_attributes(self, data)
    
        permission_overwrites = parse_permission_overwrites(data)
        if self.permission_overwrites != permission_overwrites:
            old_attributes['permission_overwrites'] = self.permission_overwrites
            self.permission_overwrites = permission_overwrites
        
        
        position = data.get('position', 0)
        if self.position != position:
            old_attributes['position'] = self.position
            self.position = position
        
        return old_attributes
    
    
    @copy_docs(ChannelMetadataGuildBase._to_data)
    def _to_data(self):
        data = ChannelMetadataGuildBase._to_data(self)
        
        data['position'] = self.position
        
        data['permission_overwrites'] = [
            {
                'id': str(permission_overwrite.target_id),
                'type': permission_overwrite.target_type.value,
                PERMISSION_ALLOW_KEY: str(permission_overwrite.allow),
                PERMISSION_DENY_KEY: str(permission_overwrite.deny)
            }
            for permission_overwrite in self.permission_overwrites.values()
        ]
        
        return data

    @classmethod
    @copy_docs(ChannelMetadataGuildBase._create_empty)
    def _create_empty(cls):
        self = super(ChannelMetadataGuildMainBase, cls)._create_empty()
        
        self.permission_overwrites = {}
        self.position = 0
        
        return self


    def _get_base_permissions_for(self, channel_entity, user):
        """
        Base permission calculator method. Subclasses call this first, then apply their channel type related changes.
        
        Parameters
        ----------
        channel_entity : ``Channel``
            The channel entity owning the metadata.
        user : ``UserBase``
            The user to calculate it's permissions of.
        
        Returns
        -------
        permissions : ``Permission``
            The calculated permissions.
        """
        guild_id = channel_entity.guild_id
        try:
            guild = GUILDS[guild_id]
        except KeyError:
            return PERMISSION_NONE
        
        if user.id == guild.owner_id:
            return PERMISSION_ALL
        
        if not isinstance(user, ClientUserBase):
            if user.guild_id in guild.channels:
                role_everyone = guild.roles.get(guild_id, None)
                if role_everyone is None:
                    permissions = PERMISSION_NONE
                else:
                    permissions = role_everyone.permissions
                
                # Apply everyone's if applicable
                permission_overwrites = self.permission_overwrites
                try:
                    permission_overwrite_everyone = permission_overwrites[guild_id]
                except KeyError:
                    pass
                else:
                    permissions &= ~permission_overwrite_everyone.deny
                    permissions |= permission_overwrite_everyone.allow
            
                return permissions
            else:
                return PERMISSION_NONE
        
        # Are we in the guild?
        try:
            guild_profile = user.guild_profiles[guild_id]
        except KeyError:
            return PERMISSION_NONE
        
        # Apply everyone's
        role_everyone = guild.roles.get(guild_id, None)
        if role_everyone is None:
            permissions = 0
        else:
            permissions = role_everyone.permissions
        
        roles = guild_profile.roles
        
        # Apply permissions of each role of the user.
        if (roles is not None):
            for role in roles:
                permissions |= role.permissions
        
        # Apply everyone's if applicable
        permission_overwrites = self.permission_overwrites
        try:
            permission_overwrite_everyone = permission_overwrites[guild_id]
        except KeyError:
            pass
        else:
            permissions &= ~permission_overwrite_everyone.deny
            permissions |= permission_overwrite_everyone.allow
        
        # Apply role overwrite
        allow = 0
        deny = 0
        
        if (roles is not None):
            for role in roles:
                try:
                    permission_overwrite_role = permission_overwrites[role.id]
                except KeyError:
                    pass
                else:
                    allow |= permission_overwrite_role.allow
                    deny |= permission_overwrite_role.deny
        
        permissions &= ~deny
        permissions |= allow
        
        # Apply user specific
        try:
            permission_overwrite_user = permission_overwrites[user.id]
        except KeyError:
            pass
        else:
            permissions &= ~permission_overwrite_user.deny
            permissions |= permission_overwrite_user.allow
        
        # Are we admin?
        if permissions & PERMISSION_MASK_ADMINISTRATOR:
            return PERMISSION_ALL
        
        return Permission(permissions)
    
    
    @copy_docs(ChannelMetadataGuildBase._get_permissions_for)
    def _get_permissions_for(self, channel_entity, user):
        result = self._get_base_permissions_for(channel_entity, user)
        if not result & PERMISSION_MASK_VIEW_CHANNEL:
            result = PERMISSION_NONE
        
        return result
    
    
    def _get_base_permissions_for_roles(self, channel_entity, roles):
        """
        Returns the channel permissions of an imaginary user who would have the listed roles. This method is called
        first by subclasses to apply their own related permissions on it.
        
        Parameters
        ----------
        channel_entity : ``Channel``
            The channel entity owning the metadata.
        roles : `tuple` of ``Role``
            The roles to calculate final permissions from.
        
        Returns
        -------
        permissions : ``Permission``
            The calculated permissions.
        
        Notes
        -----
        Partial roles and roles from other guilds as well are ignored.
        """
        guild_id = channel_entity.guild_id
        
        try:
            guild = GUILDS[guild_id]
        except KeyError:
            return PERMISSION_NONE
        
        role_everyone = guild.roles.get(guild_id, None)
        if role_everyone is None:
            permissions = 0
        else:
            permissions = role_everyone.permissions
        
        roles = sorted(roles)
        for role in roles:
            if role.guild_id == guild_id:
                permissions |= role.permissions
        
        if permissions & PERMISSION_MASK_ADMINISTRATOR:
            return PERMISSION_ALL
        
        permission_overwrites = self.permission_overwrites
        try:
            permission_overwrite_everyone = permission_overwrites[guild_id]
        except KeyError:
            pass
        else:
            permissions &= ~permission_overwrite_everyone.deny
            permissions |= permission_overwrite_everyone.allow
        
        allow = 0
        deny = 0
        
        for role in roles:
            try:
                permission_overwrite_role = permission_overwrites[role.id]
            except KeyError:
                pass
            else:
                allow |= permission_overwrite_role.allow
                deny |= permission_overwrite_role.deny
        
        permissions &= ~deny
        permissions |= allow
        
        return Permission(permissions)
    
    
    @copy_docs(ChannelMetadataGuildBase._get_permissions_for_roles)
    def _get_permissions_for_roles(self, channel_entity, roles):
        permissions = self._get_base_permissions_for_roles(channel_entity, roles)
        if not permissions & PERMISSION_MASK_VIEW_CHANNEL:
            permissions = PERMISSION_NONE
        
        return permissions



__all__ = ('AutoModerationActionMetadata', 'SendAlertMessageActionMetadata', 'TimeoutActionMetadata')

from math import ceil

from scarletio import RichAttributeErrorBaseType, copy_docs

from ..channel import Channel, create_partial_channel_from_id

from .constants import AUTO_MODERATION_ACTION_TIMEOUT_MAX


class AutoModerationActionMetadata(RichAttributeErrorBaseType):
    """
    Base class for ``AutoModerationAction``'s action metadata.
    """
    __slots__ = ()
    
    def __new__(cls):
        """
        Creates a new action metadata instance.
        """
        return object.__new__(cls)
    
    
    def __repr__(self):
        """Returns the action metadata's representation."""
        return f'<{self.__class__.__name__}>'
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new auto moderation action metadata instance from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
           Auto moderation action metadata payload.
        
        Returns
        -------
        self : ``ScheduledEventEntityMetadata``
        """
        return object.__new__(cls)
    
    
    def to_data(self):
        """
        Converts the action metadata to json serializable object.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        return {}
    
    
    def __eq__(self, other):
        """Returns whether the two action metadatas are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return True
    
    
    def __hash__(self):
        """Returns the action metadata's hash value."""
        return 0
    
    
    def copy(self):
        """
        Copies the action metadata.
        
        Returns
        -------
        new : ``AutoModerationActionMetadata``
        """
        return object.__new__(type(self))


class SendAlertMessageActionMetadata(AutoModerationActionMetadata):
    """
    Send alert message action metadata for an auto moderation action.
    
    Attributes
    ----------
    channel_id : `int`
        The channel's identifier where the alert messages are sent.
    """
    __slots__ = ('channel_id',)
    
    def __new__(cls, channel):
        """
        Creates a new send alert message action metadata.
        
        Parameters
        ----------
        channel : `None`, ``Channel``, `int`
            The channel where the alert message should be sent.
               
        Raises
        ------
        TypeError
            - If `channel`'s type is incorrect.
        """
        if channel is None:
            channel_id = 0
        
        elif isinstance(channel, Channel):
            channel_id = channel.id
        
        elif isinstance(channel, int):
            channel_id = channel
        
        else:
            raise TypeError(
                f'`channel` parameter can be `{Channel.__name__}`, `int`, '
                f'got {channel.__class__.__name__}; {channel!r}.'
            )
        
        self = object.__new__(cls)
        self.channel_id = channel_id
        return self
    
    
    @copy_docs(AutoModerationActionMetadata.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' channel_id=')
        repr_parts.append(repr(self.channel_id))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @classmethod
    @copy_docs(AutoModerationActionMetadata.from_data)
    def from_data(cls, data):
        channel_id = data.get('channel_id', None)
        if (channel_id is None):
            channel_id = 0
        
        else:
            channel_id = int(channel_id)
        
        self = object.__new__(cls)
        self.channel_id = channel_id
        return self
    
    
    @copy_docs(AutoModerationActionMetadata.to_data)
    def to_data(self):
        data = {}
        
        data['channel_id'] = self.channel_id
        
        return data
    
    
    @copy_docs(AutoModerationActionMetadata.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.channel_id != other.channel_id:
            return False
        
        return True
    
    
    @copy_docs(AutoModerationActionMetadata.__hash__)
    def __hash__(self):
        return self.channel_id
    
    
    @copy_docs(AutoModerationActionMetadata.copy)
    def copy(self):
        new = AutoModerationActionMetadata.copy(self)
        
        # channel_id
        new.channel_id = self.channel_id
        
        return new
    
    @property
    def channel(self):
        """
        Returns the channels where the alert messages are sent.
        """
        channel_id = self.channel_id
        if channel_id:
            return create_partial_channel_from_id(channel_id, -1, 0)


class TimeoutActionMetadata(AutoModerationActionMetadata):
    """
    Timeout action metadata of an auto moderation action.
    
    Attributes
    ----------
    duration : `int`
        The timeout's duration applied on trigger.
    """
    __slots__ = ('duration',)
    
    def __new__(cls, duration):
        """
        Creates a new timeout action metadata for ``AutoModerationAction``-s.
        
        Parameters
        ----------
        duration : `None`, `int`, `float`
        The timeout's duration applied on trigger.
        
        Raises
        ------
        TypeError
            - If `duration` type is incorrect.
        ValueError
            - If `duration` is out of the expected range.
        """
        if duration is None:
            duration = 0
        
        elif isinstance(duration, int):
            pass
        
        elif isinstance(duration, float):
            duration = ceil(duration)
        
        else:
            raise TypeError(
                f'`duration` can be `None`, `int`, `float`, got {duration.__class__.__name__}; {duration!r}.'
            )
        
        if duration > AUTO_MODERATION_ACTION_TIMEOUT_MAX:
            raise ValueError(
                f'`duration` can be max {AUTO_MODERATION_ACTION_TIMEOUT_MAX!r}, got {duration!r}.'
            )
        
        self = object.__new__(cls)
        self.duration = duration
        return self
    
    
    @copy_docs(AutoModerationActionMetadata.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' duration=')
        repr_parts.append(repr(self.duration))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @classmethod
    @copy_docs(AutoModerationActionMetadata.from_data)
    def from_data(cls, data):
        duration = data.get('duration_seconds', None)
        if (duration is None):
            duration = 0
        
        self = object.__new__(cls)
        self.duration = duration
        return self
    
    
    @copy_docs(AutoModerationActionMetadata.to_data)
    def to_data(self):
        data = {}
        
        data['duration_seconds'] = self.duration
        
        return data
    
    
    @copy_docs(AutoModerationActionMetadata.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.duration != other.duration:
            return False
        
        return True
    
    
    @copy_docs(AutoModerationActionMetadata.__hash__)
    def __hash__(self):
        return self.duration
    
    
    @copy_docs(AutoModerationActionMetadata.copy)
    def copy(self):
        new = AutoModerationActionMetadata.copy(self)
        
        # duration
        new.duration = self.duration
        
        return new



__all__ = ('validate_application_target_type', )

from ....discord.interaction import ApplicationCommandTargetType


DEFAULT_APPLICATION_COMMAND_TARGET_TYPE = ApplicationCommandTargetType.chat

APPLICATION_COMMAND_TARGET_TYPES_BY_NAME = {
    application_command_target_type.name: application_command_target_type for
    application_command_target_type in ApplicationCommandTargetType.INSTANCES.values()
}

APPLICATION_COMMAND_TARGET_TYPES_BY_VALUE = {
    application_command_target_type.value: application_command_target_type for
    application_command_target_type in ApplicationCommandTargetType.INSTANCES.values()
}


def validate_application_target_type(target):
    """
    Validates the given ``ApplicationCommandTargetType`` value.
    
    Parameters
    ----------
    target : `None`, `int`, `str`, ``ApplicationCommandTargetType``
        The `target` to validate.
    
    Returns
    -------
    target : ``ApplicationCommandTargetType``
        The validated `target`.
    
    Raises
    ------
    ValueError
        - If `target` could not be matched by any expected target type name or value.
    TypeError
        - If `target` is neither `None`, `int`, `str`, nor ``ApplicationCommandTargetType``.
    """
    if target is None:
        target = ApplicationCommandTargetType.none
    
    elif isinstance(target, ApplicationCommandTargetType):
        pass
    
    elif isinstance(target, str):
        if type(target) is not str:
            target = str(target)
        
        target = target.lower()
        
        try:
            target = APPLICATION_COMMAND_TARGET_TYPES_BY_NAME[target]
        except KeyError:
            raise ValueError(
                f'Unknown `target` name: {target!r}.'
            ) from None
    
    elif isinstance(target, int):
        if type(target) is not int:
            target = int(target)
        
        try:
            target = APPLICATION_COMMAND_TARGET_TYPES_BY_NAME[target]
        except KeyError:
            raise ValueError(
                f'Unknown `target` value: {target!r}.'
            ) from None
    
    else:
        raise TypeError(
            f'`target` can be `None`, `{ApplicationCommandTargetType.__name__}`, `str`,  `int`, got '
            f'{target.__class__.__name__}; {target!r}.'
        )
    
    if target is ApplicationCommandTargetType.none:
        target = DEFAULT_APPLICATION_COMMAND_TARGET_TYPE
    
    return target



__all__ = ()

from ....sticker import StickerFormat, StickerType

from ..audit_log_change import AuditLogChange

from .shared import _convert_preinstanced, convert_deprecated, convert_nothing, convert_snowflake


def convert_sticker_format(name, data):
    return _convert_preinstanced('format', data, StickerFormat)


def convert_tags(name, data):
    before = data.get('old_value', None)
    if (before is None) or (not before):
        before = None
    else:
        before = frozenset(before.split(', '))
    
    after = data.get('new_value', None)
    if (after is None) or (not after):
        after = None
    else:
        after = frozenset(after.split(', '))
    
    return AuditLogChange('tags', before, after)


def convert_sticker_type(name, data):
    return _convert_preinstanced('type', data, StickerType)


STICKER_CONVERTERS = {
    'asset': convert_deprecated,
    'available': convert_nothing,
    'description': convert_nothing,
    'format_type': convert_sticker_format,
    'guild_id': convert_snowflake,
    'id': convert_snowflake,
    'name': convert_nothing,
    'tags': convert_tags,
    'type': convert_sticker_type,
}



__all__ = ('ChannelMetadataGuildVoiceBase',)

from scarletio import copy_docs, include

from ...preconverters import preconvert_int, preconvert_preinstanced_type

from .guild_main_base import ChannelMetadataGuildMainBase


VoiceRegion = include('VoiceRegion')


class ChannelMetadataGuildVoiceBase(ChannelMetadataGuildMainBase):
    """
    Guild voice channel metadata base.
    
    Attributes
    ----------
    _permission_cache : `None`, `dict` of (`int`, ``Permission``) items
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    parent_id : `int`
        The channel's parent's identifier.
    name : `str`
        The channel's name.
    permission_overwrites : `dict` of (`int`, ``PermissionOverwrite``) items
        The channel's permission overwrites.
    position : `int`
        The channel's position.
    bitrate : `int`
        The bitrate (in bits) of the voice channel.
    region : `None`, ``VoiceRegion``
        The voice region of the channel.
    user_limit : `int`
        The maximal amount of users, who can join the voice channel, or `0` if unlimited.
    
    Class Attributes
    ----------------
    type : `int` = `-1`
        The channel's type.
    order_group: `int` = `2`
        The channel's order group used when sorting channels.
    """
    __slots__ = ('bitrate', 'region', 'user_limit')
    
    order_group = 2
    

    @copy_docs(ChannelMetadataGuildMainBase._compare_attributes_to)
    def _compare_attributes_to(self, other):
        if not ChannelMetadataGuildMainBase._compare_attributes_to(self, other):
            return False
        
        if self.bitrate != other.bitrate:
            return False
        
        if self.region is not other.region:
            return False
        
        if self.user_limit != other.user_limit:
            return False
        
        return True
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildMainBase._create_empty)
    def _create_empty(cls):
        self = super(ChannelMetadataGuildVoiceBase, cls)._create_empty()
        
        self.bitrate = 0
        self.region = None
        self.user_limit = 0
        
        return self
    
    
    @copy_docs(ChannelMetadataGuildMainBase._update_attributes)
    def _update_attributes(self, data):
        ChannelMetadataGuildMainBase._update_attributes(self, data)
        
        self.bitrate = data['bitrate']
        self.user_limit = data['user_limit']
        
        region = data.get('rtc_region', None)
        if (region is not None):
            region = VoiceRegion.get(region)
        self.region = region
    
    
    @copy_docs(ChannelMetadataGuildMainBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = ChannelMetadataGuildMainBase._difference_update_attributes(self, data)
        
        bitrate = data['bitrate']
        if self.bitrate != bitrate:
            old_attributes['bitrate'] = self.bitrate
            self.bitrate = bitrate
        
        user_limit = data['user_limit']
        if self.user_limit != user_limit:
            old_attributes['user_limit'] = self.user_limit
            self.user_limit = user_limit
        
        region = data.get('rtc_region', None)
        if (region is not None):
            region = VoiceRegion.get(region)
        if self.region is not region:
            old_attributes['region'] = self.region
            self.region = region
        
        return old_attributes
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildMainBase._precreate)
    def _precreate(cls, keyword_parameters):
        self = super(ChannelMetadataGuildVoiceBase, cls)._precreate(keyword_parameters)
        
        try:
            bitrate = keyword_parameters.pop('bitrate')
        except KeyError:
            pass
        else:
            bitrate = preconvert_int(bitrate, 'bitrate', 8000, 384000)
            self.bitrate = bitrate
            
        try:
            user_limit = keyword_parameters.pop('user_limit')
        except KeyError:
            pass
        else:
            user_limit = preconvert_int(user_limit, 'user_limit', 0, 99)
            self.user_limit = user_limit
        
        try:
            region = keyword_parameters.pop('region')
        except KeyError:
            pass
        else:
            region = preconvert_preinstanced_type(region, 'region', VoiceRegion)
            self.region = region
        
        return self
    
    
    @copy_docs(ChannelMetadataGuildMainBase._to_data)
    def _to_data(self):
        data = ChannelMetadataGuildMainBase._to_data(self)
        
        data['bitrate'] = self.bitrate
        data['user_limit'] = self.user_limit
        
        region = self.region
        if (region is not None):
            region = region.value
        
        data['rtc_region'] = region
        
        return data



__all__ = (
    'cr_pg_channel_object', 'create_partial_channel_from_data', 'create_partial_channel_from_id'
)

import reprlib, warnings
from datetime import datetime

from scarletio import export, include

from ..bases import maybe_snowflake
from ..core import CHANNELS
from ..permission import PermissionOverwrite
from ..utils import datetime_to_timestamp

from . import channel_types as CHANNEL_TYPES
from .channel_types import get_channel_type_name, get_channel_type_names
from .constants import AUTO_ARCHIVE_OPTIONS
from .deprecation import ChannelBase
from .preinstanced import VideoQualityMode


Channel = include('Channel')
Guild = include('Guild')
VoiceRegion = include('VoiceRegion')


def create_partial_channel_from_data(data, guild_id):
    """
    Creates a partial channel from partial channel data.
    
    Parameters
    ----------
    data : `None`, `dict` of (`str`, `Any`) items
        Partial channel data received from Discord.
    guild_id : `int`
        The channel's guild's identifier.
    
    Returns
    -------
    channel : `None`, ``Channel``
        The created partial channel, or `None`, if no data was received.
    """
    if (data is None) or (not data):
        return None
    
    channel_id = int(data['id'])
    try:
        return CHANNELS[channel_id]
    except KeyError:
        pass
    
    channel = Channel._from_partial_data(data, channel_id, guild_id)
    CHANNELS[channel_id] = channel
    
    return channel


@export
def create_partial_channel_from_id(channel_id, channel_type, guild_id):
    """
    Creates a new partial channel from the given identifier.
    
    Parameters
    ----------
    channel_id : `int`
        The channel's identifier.
    channel_type : `int`
        The channel's type identifier.
    guild_id : `int`
        A guild's identifier of the created channel.
    """
    try:
        return CHANNELS[channel_id]
    except KeyError:
        pass
    
    channel = Channel._create_empty(channel_id, channel_type, guild_id)
    CHANNELS[channel_id] = channel
    
    return channel


HAS_PERMISSION_OVERWRITES = (
    CHANNEL_TYPES.guild_text,
    CHANNEL_TYPES.guild_voice,
    CHANNEL_TYPES.guild_category,
    CHANNEL_TYPES.guild_announcements,
    CHANNEL_TYPES.guild_store,
    CHANNEL_TYPES.guild_stage,
    CHANNEL_TYPES.guild_directory,
    CHANNEL_TYPES.guild_forum,
)

HAS_TOPIC = (
    CHANNEL_TYPES.guild_text,
    CHANNEL_TYPES.guild_announcements,
    CHANNEL_TYPES.guild_stage,
    CHANNEL_TYPES.guild_forum,
)

HAS_NSFW = (
    *CHANNEL_TYPES.GROUP_GUILD_MAIN_TEXT,
    CHANNEL_TYPES.guild_voice,
)

HAS_SLOWMODE = (
    *CHANNEL_TYPES.GROUP_MESSAGEABLE,
    CHANNEL_TYPES.guild_forum,
    *CHANNEL_TYPES.GROUP_THREAD,
)

HAS_BITRATE = CHANNEL_TYPES.GROUP_GUILD_CONNECTABLE

HAS_USER_LIMIT = CHANNEL_TYPES.GROUP_GUILD_CONNECTABLE

HAS_REGION = CHANNEL_TYPES.GROUP_GUILD_CONNECTABLE

HAS_VIDEO_QUALITY_MODE = (
    CHANNEL_TYPES.guild_voice,
)

HAS_ARCHIVED = CHANNEL_TYPES.GROUP_THREAD

HAS_ARCHIVED_AT = CHANNEL_TYPES.GROUP_THREAD

HAS_AUTO_ARCHIVE_AFTER = CHANNEL_TYPES.GROUP_THREAD

HAS_DEFAULT_AUTO_ARCHIVE_AFTER = CHANNEL_TYPES.GROUP_CAN_CONTAIN_THREADS

HAS_OPEN = CHANNEL_TYPES.GROUP_THREAD

CAN_HAVE_PARENT_ID = (
    CHANNEL_TYPES.guild_category,
    *CHANNEL_TYPES.GROUP_CAN_CONTAIN_THREADS,
)


def cr_pg_channel_object(
    name, channel_type, *, permission_overwrites=..., topic=..., nsfw=..., slowmode=..., bitrate=..., user_limit=...,
    region=..., video_quality_mode=..., archived=..., archived_at=..., auto_archive_after=..., open_=...,
    default_auto_archive_after=..., parent=..., guild=None,
):
    """
    Creates a json serializable object representing a ``Channel``.
    
    Parameters
    ----------
    name : `str`
        The name of the channel. Can be between `1` and `100` characters.
    channel_type : `int`
        The channel's type.
    permission_overwrites : `None`, `list` of ``cr_p_permission_overwrite_object`` returns, Optional (Keyword only)
        A list of permission overwrites of the channel. The list should contain json serializable permission
        overwrites made by the ``cr_p_permission_overwrite_object`` function.
    topic : `None`, `str`, Optional (Keyword only)
        The channel's topic.
    nsfw : `None`, `bool`, Optional (Keyword only)
        Whether the channel is marked as nsfw.
    slowmode : `None`, `int`, Optional (Keyword only)
        The channel's slowmode value.
    bitrate : `None`, `int`, Optional (Keyword only)
        The channel's bitrate.
    user_limit : `int`, Optional (Keyword only)
        The channel's user limit.
    region : `None`, ``VoiceRegion``, `str`, Optional (Keyword only)
        The channel's voice region.
    video_quality_mode : `None`, ``VideoQualityMode``, `int`, Optional (Keyword only)
        The channel's video quality mode.
    archived : `None`, `bool`, Optional (Keyword only)
        Whether the thread channel is archived.
    archived_at : `None`, `datetime`, Optional (Keyword only)
        When the thread's archive status was last changed.
    auto_archive_after : `None`, `int`, Optional (Keyword only)
        Duration in minutes to automatically archive the thread after recent activity. Can be one of: `3600`, `86400`,
        `259200`, `604800`.
    open_ : `None`, `bool`, Optional (Keyword only)
        Whether the thread channel is open.
    default_auto_archive_after : `None`, `int`, Optional (Keyword only)
        The default duration (in seconds) for newly created threads to automatically archive the themselves. Can be
        one of: `3600`, `86400`, `259200`, `604800`.
    parent : `None`, ``Channel``, `int`, Optional (Keyword only)
        The channel's parent. If the parent is under a guild, leave it empty.
    category : `None`, ``Channel``, `int`, Optional (Keyword only)
        Deprecated, please use `parent` parameter instead.
    guild : `None`, ``Guild`` = `None`, Optional (Keyword only)
        Reference guild used for validation purposes. Defaults to `None`.
    
    Returns
    -------
    channel_data : `dict` of (`str`, `Any`) items
    
    Raises
    ------
    TypeError
        - If `channel_type` was not passed as `int`
        - If `parent` was not given as `None`, ``Channel``, `int`.
        - If `region` was not given either as `None`, `str` nor ``VoiceRegion``.
        - If `video_quality_mode` was not given neither as `None`, `VideoQualityMode`` nor as `int`.
    AssertionError
        - If any parameter's type or value is incorrect.
    """
    if __debug__:
        if (guild is not None) and (not isinstance(guild, Guild)):
            raise AssertionError(
                '`guild` is given, but not as `None` nor `Guild`, got '
                f'{guild.__class__.__name__}; {guild!r}.'
            )
    
    if isinstance(channel_type, int):
        if __debug__:
            if channel_type not in CHANNEL_TYPES.GROUP_IN_PRODUCTION:
                raise AssertionError(
                    f'`channel_type` is not in any of the in-production channel types: '
                    f'{CHANNEL_TYPES.GROUP_IN_PRODUCTION}, got {channel_type!r}.'
                )
    else:
        if isinstance(channel_type, ChannelBase):
            warnings.warn(
                (
                    f'`cr_pg_channel_object`\'s `channel_type` parameter cannot be `{ChannelBase.__name__}` '
                    f'subclass, got {channel_type.__class__.__name__}; {channel_type!r}.'
                ),
                FutureWarning,
                stacklevel = 2,
            )
            
            channel_type = channel_type.type
    
    
    if channel_type not in CHANNEL_TYPES.GROUP_GUILD:
        raise TypeError(
            f'`channel_type` not refers to a guild channel type. Got {channel_type!r}.'
        )
    
    
    if not isinstance(name, str):
        raise AssertionError(
            f'`name` can be `str`, got {name.__class__.__name__}; {name!r}.'
        )
    
    name_length = len(name)
    
    if name_length < 1 or name_length > 100:
        raise AssertionError(
            f'`name` length can be in range [1:100], got {name_length}; {name!r}.'
        )
    
    channel_data = {
        'name': name,
        'type': channel_type,
    }
    
    if channel_type in HAS_PERMISSION_OVERWRITES:
        if permission_overwrites is ...:
            permission_overwrites = []
        else:
            if __debug__:
                if not isinstance(permission_overwrites, list):
                    raise AssertionError(
                        f'`permission_overwrites` can be `None`, `list` of '
                        f'`cr_p_permission_overwrite_object` returns, got {permission_overwrites.__class__.__name__}; '
                        f'{permission_overwrites!r}.'
                    )
                
                for index, element in enumerate(permission_overwrites):
                    if not isinstance(element, dict):
                        raise AssertionError(
                            f'`permission_overwrites`\'s element {index} should be `dict`, '
                            f'but got {element.__class__.__name__}; {element!r}.'
                        )
        
        channel_data['permission_overwrites'] = permission_overwrites
    else:
        if __debug__:
            if (permission_overwrites is not ...):
                _assert_channel_type(
                    channel_type, None, HAS_PERMISSION_OVERWRITES, 'permission_overwrites', permission_overwrites,
                )
    
    
    _maybe_add_channel_topic_field_to_data(channel_type, None, channel_data, topic)
    _maybe_add_channel_nsfw_field_to_data(channel_type, None, channel_data, nsfw)
    _maybe_add_channel_slowmode_field_to_data(channel_type, None, channel_data, slowmode)
    _maybe_add_channel_bitrate_field_to_data(channel_type, None, channel_data, bitrate)
    _maybe_add_channel_user_limit_field_to_data(channel_type, None, channel_data, user_limit)
    _maybe_add_channel_region_field_to_data(channel_type, None, channel_data, region)
    _maybe_add_channel_video_quality_mode_field_to_data(channel_type, None, channel_data, video_quality_mode)
    
    if (archived is not ...):
        if __debug__:
            _assert_channel_type(channel_type, None, HAS_ARCHIVED, 'archived', archived)
            
            if not isinstance(archived, bool):
                raise AssertionError(
                    f'`archived` can be `None`, `bool`, got {archived.__class__.__name__}; {archived!r}.'
                )
        
        channel_data['archived'] = archived
    
    
    if (archived_at is not ...):
        if __debug__:
            _assert_channel_type(channel_type, None, HAS_ARCHIVED_AT, 'archived_at', archived_at)
            
            if not isinstance(archived_at, datetime):
                raise AssertionError(
                    f'`archived_at` can be `None`, `datetime`, got {archived_at.__class__.__name__}; {archived_at!r}.'
                )
        
        channel_data['archive_timestamp'] =  datetime_to_timestamp(archived_at)
    
    
    if (auto_archive_after is not ...):
        if __debug__:
            _assert_channel_type(channel_type, None, HAS_AUTO_ARCHIVE_AFTER, 'auto_archive_after', auto_archive_after)
            
            if not isinstance(auto_archive_after, int):
                raise AssertionError(
                    f'`auto_archive_after` can be `None`, `datetime`, got {auto_archive_after.__class__.__name__}; '
                    f'{auto_archive_after!r}.'
                )
            
            if auto_archive_after not in AUTO_ARCHIVE_OPTIONS:
                raise AssertionError(
                    f'`auto_archive_after` can be any of: {AUTO_ARCHIVE_OPTIONS}, got {auto_archive_after}.'
                )
        
        channel_data['auto_archive_duration'] = auto_archive_after // 60
    
    
    _maybe_add_channel_default_auto_archive_after_field_to_data(
        channel_type, None, channel_data, default_auto_archive_after
    )
    
    
    if (open_ is not ...):
        if __debug__:
            _assert_channel_type(channel_type, None, HAS_OPEN, 'open_', open_)
            
            if not isinstance(open_, bool):
                raise AssertionError(
                    f'`open_` can be `None`, `bool`, got {open_.__class__.__name__}; {open_!r}.'
                )
        
        channel_data['locked'] = not open_
    
    
    if parent is None:
        parent_id = 0
    
    elif isinstance(parent, Channel):
        parent_id = parent.id
    
    else:
        parent_id = maybe_snowflake(parent)
        if parent_id is None:
            raise TypeError(
                f'`parent` can be `{Channel.__name__}`, `{Guild.__name__}`, `int`, got '
                f'{parent.__class__.__name__}; {parent!r}.'
            )
    
    if parent_id:
        if __debug__:
            if channel_type != CAN_HAVE_PARENT_ID:
                raise AssertionError(
                    f'`parent_id` is only applicable for '
                    f'{get_channel_type_names(HAS_OPEN)} channels'
                    f', got {get_channel_type_name(channel_type)}; {channel_type!r}.'
                )
        
        channel_data['parent_id'] = parent_id
    
    
    return channel_data


def _assert_channel_type(channel_type, channel, accepted_types, field_name, field_value):
    """
    Asserts whether the given `channel_type`, `channel` combination passes the `accepted_types` check.
    
    Parameters
    ----------
    channel_type : `int`
        The channel's type.
    channel : `None`, ``Channel``
        The received channel type if any.
    accepted_types : `iterable` of `int`
        The accepted channel types.
    field_name : `str`
        The field's name to assert.
    field_value : `Any`
        The field's value.
    
    Raises
    ------
    AssertionError
    """
    if (channel_type in accepted_types):
        return
    
    if (channel is not None) and channel.partial:
        return
    
    _raise_channel_type_assertion_with_message(channel_type, channel, accepted_types, field_name, field_value)


def _raise_channel_type_assertion_with_message(channel_type, channel, accepted_types, field_name, field_value):
    """
    Builds a channel type assertion message and raises assertion error.
    
    Parameters
    ----------
    channel_type : `int`
        The channel's type.
    channel : `None`, ``Channel``
        The received channel type if any.
    accepted_types : `iterable` of `int`
        The accepted channel types.
    field_name : `str`
        The field's name to assert.
    field_value : `Any`
        The field's value.
    
    Raises
    ------
    AssertionError
    """
    assertion_message_parts = []
    
    assertion_message_parts.append('`')
    assertion_message_parts.append(field_name)
    assertion_message_parts.append('` is only valid for ')
    assertion_message_parts.append(get_channel_type_names(accepted_types))
    assertion_message_parts.append(', got ')
    assertion_message_parts.append(get_channel_type_name(channel_type))
    
    if channel is not None:
        assertion_message_parts.append('; channel=')
        assertion_message_parts.append(repr(channel))
    
    assertion_message_parts.append('; ')
    assertion_message_parts.append(field_name)
    assertion_message_parts.append('=˙')
    assertion_message_parts.append(reprlib.repr(field_value))
    assertion_message_parts.append('˙.')
    
    assertion_message = ''.join(assertion_message_parts)
    assertion_message_parts = None
    
    raise AssertionError(assertion_message)


def _maybe_add_channel_bitrate_field_to_data(channel_type, channel, channel_data, bitrate):
    """
    Adds the given `bitrate` field to `channel_data` if the value is not ellipsis.
    
    Parameters
    ----------
    channel_type : `int`
        The respective channel's type.
    channel_data : `dict` of (`str`, `Any`) items
        Channel data to add the field to.
    bitrate : `Ellipsis`, `int`
        `channel.bitrate` field.
    
    Raises
    ------
    AssertionError
        - If `bitrate`'s type or value is incorrect.
    """
    if (bitrate is not ...):
        if __debug__:
            _assert_channel_type(channel_type, channel, HAS_BITRATE, 'bitrate', bitrate)
            
            if not isinstance(bitrate, int):
                raise AssertionError(
                    f'`bitrate` can be `int`, got {bitrate.__class__.__name__}; {bitrate!r}.'
                )
            
            # Get max bitrate
            if guild is None:
                bitrate_limit = 384000
            else:
                bitrate_limit = guild.bitrate_limit
            
            if bitrate < 8000 or bitrate > bitrate_limit:
                raise AssertionError(
                    f'`bitrate` is out of the expected [8000:{bitrate_limit}] range, got {bitrate!r}.'
                )
        
        channel_data['bitrate'] = bitrate


def _maybe_add_channel_topic_field_to_data(channel_type, channel, channel_data, topic):
    """
    Adds the given `topic` field to `channel_data` if the value is not ellipsis.
    
    Parameters
    ----------
    channel_type : `int`
        The respective channel's type.
    channel_data : `dict` of (`str`, `Any`) items
        Channel data to add the field to.
    topic : `Ellipsis`, `str`
        `channel.topic` field.
    
    Raises
    ------
    AssertionError
        - If `topic`'s type or value is incorrect.
    """
    if (topic is not ...):
        if __debug__:
            _assert_channel_type(channel_type, channel, HAS_TOPIC, 'topic', topic)
            
            if not isinstance(topic, str):
                raise AssertionError(
                    f'`topic` can be `str`, got {topic.__class__.__name__}; {topic!r}.'
                )
            
            if channel_type in (
                CHANNEL_TYPES.guild_text,
                CHANNEL_TYPES.guild_announcements,
            ):
                topic_length_limit = 1024
            else:
                topic_length_limit = 120
            
            topic_length = len(topic)
            if topic_length > topic_length_limit:
                raise AssertionError(
                    f'`topic` length can be in range [0:{topic_length_limit}], got {topic_length}; {topic!r}.'
                )
        
        channel_data['topic'] = topic


def _maybe_add_channel_nsfw_field_to_data(channel_type, channel, channel_data, nsfw):
    """
    Adds the given `nsfw` field to `channel_data` if the value is not ellipsis.
    
    Parameters
    ----------
    channel_type : `int`
        The respective channel's type.
    channel_data : `dict` of (`str`, `Any`) items
        Channel data to add the field to.
    nsfw : `Ellipsis`, `bool`
        `channel.nsfw` field.
    
    Raises
    ------
    AssertionError
        - If `nsfw`'s type or value is incorrect.
    """
    if (nsfw is not ...):
        if __debug__:
            _assert_channel_type(channel_type, channel, HAS_NSFW, 'nsfw', nsfw)
            
            if not isinstance(nsfw, bool):
                raise AssertionError(
                    f'`nsfw` can be `bool`, got {nsfw.__class__.__name__}; {nsfw!r}.'
                )
        
        channel_data['nsfw'] = nsfw
        

def _maybe_add_channel_slowmode_field_to_data(channel_type, channel, channel_data, slowmode):
    """
    Adds the given `slowmode` field to `channel_data` if the value is not ellipsis.
    
    Parameters
    ----------
    channel_type : `int`
        The respective channel's type.
    channel_data : `dict` of (`str`, `Any`) items
        Channel data to add the field to.
    slowmode : `Ellipsis`, `int`
        `channel.slowmode` field.
    
    Raises
    ------
    AssertionError
        - If `slowmode`'s type or value is incorrect.
    """
    if (slowmode is not ...):
        if __debug__:
            _assert_channel_type(channel_type, channel, HAS_SLOWMODE, 'slowmode', slowmode)
            
            if not isinstance(slowmode, int):
                raise AssertionError(
                    f'`slowmode` can be `int`, got {slowmode.__class__.__name__}; {slowmode!r}.'
                )
            
            if slowmode < 0 or slowmode > 21600:
                raise AssertionError(
                    f'`slowmode` can be in range [0:21600], got: {slowmode!r}.'
                )
        
        channel_data['rate_limit_per_user'] = slowmode


def _maybe_add_channel_user_limit_field_to_data(channel_type, channel, channel_data, user_limit):
    """
    Adds the given `user_limit` field to `channel_data` if the value is not ellipsis.
    
    Parameters
    ----------
    channel_type : `int`
        The respective channel's type.
    channel_data : `dict` of (`str`, `Any`) items
        Channel data to add the field to.
    user_limit : `Ellipsis`, `int`
        `channel.user_limit` field.
    
    Raises
    ------
    AssertionError
        - If `user_limit`'s type or value is incorrect.
    """
    if (user_limit is not ...):
        if __debug__:
            _assert_channel_type(channel_type, channel, HAS_USER_LIMIT, 'user_limit', user_limit)
            
            if not isinstance(user_limit, int):
                raise AssertionError(
                    f'`user_limit` can be `int`, got {user_limit.__class__.__name__}; '
                    f'{user_limit!r}.'
                )
            
            if user_limit < 0 or user_limit > 99:
                raise AssertionError(
                    '`user_limit`\'s value is out of the expected [0:99] range, got '
                    f'{user_limit!r}.'
                )
        
        channel_data['user_limit'] = user_limit


def _maybe_add_channel_region_field_to_data(channel_type, channel, channel_data, region):
    """
    Adds the given `region` field to `channel_data` if the value is not ellipsis.
    
    Parameters
    ----------
    channel_type : `int`
        The respective channel's type.
    channel_data : `dict` of (`str`, `Any`) items
        Channel data to add the field to.
    region : `Ellipsis`, `None`, `str`, ``VoiceRegion``
        `channel.region` field.
    
    Raises
    ------
    AssertionError
        - If `region`'s type or value is incorrect.
    TypeError
        - If `region`'s type is incorrect.
    """
    if (region is not ...):
        if __debug__:
            _assert_channel_type(channel_type, channel, HAS_REGION, 'region', region)
        
        if (region is not None):
            region_value = None
        
        elif isinstance(region, VoiceRegion):
            region_value = region.value
        
        elif isinstance(region, str):
            region_value = region
        
        else:
            raise TypeError(
                f'`region` can be `None`, `str`, `{VoiceRegion.__name__}`, got '
                f'{region.__class__.__name__}; {region!r}.'
            )
        
        channel_data['rtc_region'] = region_value


def _maybe_add_channel_video_quality_mode_field_to_data(channel_type, channel, channel_data, video_quality_mode):
    """
    Adds the given `video_quality_mode` field to `channel_data` if the value is not ellipsis.
    
    Parameters
    ----------
    channel_type : `int`
        The respective channel's type.
    channel_data : `dict` of (`str`, `Any`) items
        Channel data to add the field to.
    video_quality_mode : `Ellipsis`, ``VideoQualityMode``, `int`
        `channel.video_quality_mode` field.
    
    Raises
    ------
    AssertionError
        - If `video_quality_mode`'s type or value is incorrect.
    TypeError
        - If `video_quality_mode`'s type is incorrect.
    """
    if (video_quality_mode is not ...):
        if __debug__:
            _assert_channel_type(
                channel_type, channel, HAS_VIDEO_QUALITY_MODE, 'video_quality_mode', video_quality_mode
            )
        
        if isinstance(video_quality_mode, VideoQualityMode):
            video_quality_mode_value = video_quality_mode.value
        
        elif isinstance(video_quality_mode, int):
            video_quality_mode_value = video_quality_mode
        
        else:
            raise TypeError(
                f'`video_quality_mode` can be `None`, `str`, `{VideoQualityMode.__name__}`, got '
                f'{video_quality_mode.__class__.__name__}; {video_quality_mode!r}.'
            )
        
        channel_data['video_quality_mode'] = video_quality_mode_value


def _maybe_add_channel_default_auto_archive_after_field_to_data(
    channel_type, channel, channel_data, default_auto_archive_after
):
    """
    Adds the given `default_auto_archive_after` field to `channel_data` if the value is not ellipsis.
    
    Parameters
    ----------
    channel_type : `int`
        The respective channel's type.
    channel_data : `dict` of (`str`, `Any`) items
        Channel data to add the field to.
    default_auto_archive_after : `Ellipsis`, `int`
        `channel.default_auto_archive_after` field.
    
    Raises
    ------
    AssertionError
        - If `default_auto_archive_after`'s type or value is incorrect.
    """
    if (default_auto_archive_after is not ...):
        if __debug__:
            _assert_channel_type(
                channel_type,
                channel,
                HAS_DEFAULT_AUTO_ARCHIVE_AFTER,
                'default_auto_archive_after',
                default_auto_archive_after,
            )
            
            if not isinstance(default_auto_archive_after, int):
                raise AssertionError(
                    f'`default_auto_archive_after` can be `None`, `datetime`, got '
                    f'{default_auto_archive_after.__class__.__name__}; {default_auto_archive_after!r}.'
                )
            
            if default_auto_archive_after not in AUTO_ARCHIVE_OPTIONS:
                raise AssertionError(
                    f'`default_auto_archive_after` can be any of: {AUTO_ARCHIVE_OPTIONS}, got '
                    f'{default_auto_archive_after}.'
                )
        
        channel_data['default_auto_archive_duration'] = default_auto_archive_after // 60


@export
def parse_permission_overwrites(data):
    """
    Parses the permission overwrites from the given data and returns them.
    
    Parameters
    ----------
    data : `list` of (`dict` of (`str`, `Any`) items) elements
        A list of permission overwrites' data.
    
    Returns
    -------
    permission_overwrites : `dict` of (`int`, ``PermissionOverwrite``) items
    """
    permission_overwrites = {}
    
    permission_overwrites_datas = data.get('permission_overwrites', None)
    if (permission_overwrites_datas is not None) and permission_overwrites_datas:
        for permission_overwrite_data in permission_overwrites_datas:
            permission_overwrite = PermissionOverwrite(permission_overwrite_data)
            permission_overwrites[permission_overwrite.target_id] = permission_overwrite
    
    return permission_overwrites



__all__ = ('ChannelMetadataGuildThreadBase', )

from datetime import datetime

from scarletio import copy_docs

from ...core import CHANNELS, GUILDS
from ...permission.permission import PERMISSION_NONE
from ...preconverters import preconvert_bool, preconvert_int, preconvert_int_options
from ...utils import datetime_to_timestamp, id_to_datetime, timestamp_to_datetime

from ..constants import AUTO_ARCHIVE_DEFAULT, AUTO_ARCHIVE_OPTIONS

from .guild_base import ChannelMetadataGuildBase


class ChannelMetadataGuildThreadBase(ChannelMetadataGuildBase):
    """
    Base guild channel metadata type.
    
    Attributes
    ----------
    _permission_cache : `None`, `dict` of (`int`, ``Permission``) items
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    parent_id : `int`
        The channel's parent's identifier.
    name : `str`
        The channel's name.
    _created_at : `None`, `datetime`
        When the channel was created.
    archived : `bool`
        Whether the thread s archived.
    archived_at : `None`, `datetime`
        When the thread's archive status was last changed.
    auto_archive_after : `int`
        Duration in seconds to automatically archive the thread after recent activity. Can be one of: `3600`, `86400`,
        `259200`, `604800`.
    open : `bool`
        Whether the thread channel is open.
    slowmode : `int`
        The amount of time in seconds what a user needs to wait between it's each message. Bots and user accounts with
        `manage_messages`, `manage_channel` permissions are unaffected.
    thread_users : `None`, `dict` of (`int`, ``ClientUserBase``) items
        The users inside of the thread if any.
    owner_id : `int`
        The channel's creator's identifier. Defaults to `0`.
    
    Class Attributes
    ----------------
    type : `int` = `-1`
        The channel's type.
    order_group: `int` = `0`
        The channel's order group used when sorting channels.
    """
    __slots__ = (
        '_created_at', 'archived', 'archived_at', 'auto_archive_after', 'open', 'owner_id', 'slowmode', 'thread_users'
    )
    
    @copy_docs(ChannelMetadataGuildBase.__new__)
    def __new__(cls, data):
        self = ChannelMetadataGuildBase.__new__(cls, data)
        
        owner_id = data.get('owner_id', None)
        if owner_id is None:
            owner_id = 0
        else:
            owner_id = int(owner_id)
        self.owner_id = owner_id
        
        created_at = data.get('create_timestamp', None)
        if (created_at is not None):
            created_at = timestamp_to_datetime(created_at)
        self._created_at = created_at
        
        
        parent_id = data.get('parent_id', None)
        if (parent_id is None):
            parent_id = 0
        else:
            parent_id = int(parent_id)
        self.parent_id = parent_id
        
        return self
    
    
    @copy_docs(ChannelMetadataGuildBase._created)
    def _created(self, channel_entity, client):
        try:
            guild = GUILDS[channel_entity.guild_id]
        except KeyError:
            pass
        else:
            guild.threads[channel_entity.id] = channel_entity
    
    
    @copy_docs(ChannelMetadataGuildBase._compare_attributes_to)
    def _compare_attributes_to(self, other):
        if not ChannelMetadataGuildBase._compare_attributes_to(self, other):
            return False
        
        if self._created_at != other._created_at:
            return False
        
        if self.archived != other.archived:
            return False
        
        if self.archived_at != other.archived_at:
            return False
        
        if self.auto_archive_after != other.auto_archive_after:
            return False
        
        if self.open != other.open:
            return False
        
        if self.owner_id != other.owner_id:
            return False
        
        if self.slowmode != other.slowmode:
            return False
        
        # Ignoring `thread_users`
        
        return True
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildBase._create_empty)
    def _create_empty(cls):
        self = super(ChannelMetadataGuildThreadBase, cls)._create_empty()
        
        self._created_at = None
        self.archived = False
        self.archived_at = None
        self.auto_archive_after = AUTO_ARCHIVE_DEFAULT
        self.open = False
        self.owner_id = 0
        self.slowmode = 0
        self.thread_users = None
        
        return self
    
    
    @copy_docs(ChannelMetadataGuildBase._get_display_name)
    def _get_display_name(self):
        return self.name.lower()
    
    
    @copy_docs(ChannelMetadataGuildBase._get_created_at)
    def _get_created_at(self, channel_entity):
        created_at = self._created_at
        if (created_at is None):
            created_at = id_to_datetime(channel_entity.id)
        
        return created_at
    
    
    @property
    @copy_docs(ChannelMetadataGuildBase._get_users)
    def _get_users(self, channel_entity):
        thread_users = self.thread_users
        if thread_users is None:
            users = []
        else:
            users = list(thread_users.values())
        
        return users
    
    
    @copy_docs(ChannelMetadataGuildBase._iter_users)
    def _iter_users(self):
        thread_users = self.thread_users
        if (thread_users is not None):
            yield from thread_users.values()
    
    
    @copy_docs(ChannelMetadataGuildBase._update_attributes)
    def _update_attributes(self, data):
        ChannelMetadataGuildBase._update_attributes(self, data)
        
        slowmode = data.get('rate_limit_per_user', None)
        if slowmode is None:
            slowmode = 0
        self.slowmode = slowmode
        
        # Move to sub data
        data = data['thread_metadata']
        
        self.archived = data.get('archived', False)
        
        
        self.auto_archive_after = data['auto_archive_duration'] * 60
        
        archived_at_data = data.get('archive_timestamp', None)
        if archived_at_data is None:
            archived_at = None
        else:
            archived_at = timestamp_to_datetime(archived_at_data)
        self.archived_at = archived_at
        
        self.open = not data.get('locked', True)
    
    
    @copy_docs(ChannelMetadataGuildBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = ChannelMetadataGuildBase._difference_update_attributes(self, data)
        
        slowmode = data.get('rate_limit_per_user', None)
        if slowmode is None:
            slowmode = 0
        if self.slowmode != slowmode:
            old_attributes['slowmode'] = self.slowmode
            self.slowmode = slowmode
        
        
        # Move to sub data
        data = data['thread_metadata']
        
        
        archived = data.get('archived', False)
        if (self.archived != archived):
            old_attributes['archived'] = self.archived
            self.archived = archived
        
        
        auto_archive_after = data['auto_archive_duration']*60
        if (self.auto_archive_after != auto_archive_after):
            old_attributes['auto_archive_after'] = self.auto_archive_after
            self.auto_archive_after = auto_archive_after
        
        
        archived_at_data = data.get('archive_timestamp', None)
        if archived_at_data is None:
            archived_at = None
        else:
            archived_at = timestamp_to_datetime(archived_at_data)
        if (self.archived_at != archived_at):
            old_attributes['archived_at'] = self.archived_at
            self.archived_at = archived_at
        
        open_ = not data.get('locked', True)
        if (self.open != open_):
            old_attributes['open'] = self.open
            self.open = open_
        
        return old_attributes
    
    
    @copy_docs(ChannelMetadataGuildBase._delete)
    def _delete(self, channel_entity, client):
        guild_id = channel_entity.guild_id
        try:
            guild = GUILDS[guild_id]
        except KeyError:
            pass
        else:
            try:
                del guild.threads[channel_entity.id]
            except KeyError:
                pass
        
        thread_users = self.thread_users
        if (thread_users is not None):
            self.thread_users = None
            channel_id = channel_entity.id
            for user in thread_users.values():
                thread_profiles = user.thread_profiles
                if (thread_profiles is not None):
                    try:
                        del thread_profiles[channel_id]
                    except KeyError:
                        pass
                    else:
                        if (not thread_profiles):
                            user.thread_profiles = None
    
    
    @copy_docs(ChannelMetadataGuildBase._get_permissions_for)
    def _get_permissions_for(self, channel_entity, user):
        try:
            parent = CHANNELS[self.parent_id]
        except KeyError:
            return PERMISSION_NONE
        
        return parent.permissions_for(user)
    
    
    @copy_docs(ChannelMetadataGuildBase._get_permissions_for_roles)
    def _get_permissions_for_roles(self, channel_entity, roles):
        try:
            parent = CHANNELS[self.parent_id]
        except KeyError:
            return PERMISSION_NONE
        
        return parent.permissions_for_roles(*roles)
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildBase._precreate)
    def _precreate(cls, keyword_parameters):
        self = super(ChannelMetadataGuildThreadBase, cls)._precreate(keyword_parameters)
    
        try:
            slowmode = keyword_parameters.pop('slowmode')
        except KeyError:
            pass
        else:
            slowmode = preconvert_int(slowmode, 'slowmode', 0, 21600)
            self.slowmode = slowmode
        
        
        try:
            auto_archive_after = keyword_parameters.pop('auto_archive_after')
        except KeyError:
            pass
        else:
            auto_archive_after = preconvert_int_options(
                auto_archive_after,
                'auto_archive_after',
                AUTO_ARCHIVE_OPTIONS,
            )
            
            self.auto_archive_after = auto_archive_after
        
        
        try:
            created_at = keyword_parameters.pop('created_at')
        except KeyError:
            pass
        else:
            if (created_at is not None):
                if not isinstance(created_at, datetime):
                    raise TypeError(
                        f'`created_at` can be `None`, `datetime`, got {created_at.__class__.__name__}; '
                        f'{created_at!r}.'
                    )
                
                self._created_at = created_at
        
        try:
            open_ = keyword_parameters.pop('open')
        except KeyError:
            pass
        else:
            open_ = preconvert_bool(open_, 'open')
            self.open_ = open_
        
        
        return self
    
    
    @copy_docs(ChannelMetadataGuildBase._to_data)
    def _to_data(self):
        data = ChannelMetadataGuildBase._to_data(self)
        
        slowmode = self.slowmode
        if slowmode:
            data['rate_limit_per_user'] = slowmode
        
        owner_id = self.owner_id
        if owner_id:
            data['owner_id'] = str(owner_id)
        
        thread_data = {}
        data['thread_metadata'] = thread_data
        
        thread_data['archived'] = self.archived
        
        thread_data['auto_archive_duration'] = self.auto_archive_after // 60
        
        archive_timestamp = self.archive_timestamp
        if (archive_timestamp is not None):
            archive_timestamp = datetime_to_timestamp(archive_timestamp)
        thread_data['archive_timestamp'] = archive_timestamp
        
        thread_data['locked'] = not self.open
        
        created_at = self._created_at
        if (created_at is not None):
            data['create_timestamp'] = datetime_to_timestamp(created_at)
        
        return data



__all__ = ('AuditLogEntry', )

from ....env import ALLOW_DEBUG_MESSAGES
from ....utils.debug import call_debug_logger

from ...utils import id_to_datetime

from .change_converters.all_ import MERGED_CONVERTERS
from .change_converters.shared import convert_nothing
from .detail_converters import DETAIL_CONVERTERS, detail_converter_nothing
from .preinstanced import AuditLogEvent


class AuditLogEntry:
    """
    Represents an entry of an ``AuditLog``.
    
    Attributes
    ----------
    _parent_reference : ``WeakReferer`` to ``AuditLog``
        Reference to the audit log entry keys' parent.
    changes : `list` of ``AuditLogChange``
        The changes of the entry.
    details : `None`, `dict` of (l`str`, `Any`) items
        Additional information for a specific action types.
    id : `int`
        The unique identifier number of the entry.
    reason : `None`, `str`
        The reason provided with the logged action.
    target_id : `int`
        The entry's target's identifier.
        
        If no target is provided, this value defaults to `0`.
    type : ``AuditLogEvent``
        The event type of the logged action.
    user : `None`, ``ClientUserBase``
        The user, who executed the logged action. If no user is provided then can be `None` as well.
    """
    __slots__ = ('_parent_reference', 'changes', 'details', 'id', 'reason', 'target_id', 'type', 'user',)
    
    def __new__(cls, data, parent):
        """
        Creates an audit log entry, from entry data sent inside of an ``AuditLog``'s data and from the audit itself.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Data received from Discord.
        parent : ``AuditLog``
            The parent of the entry, what contains the respective guild, the included users, webhooks and the
            integrations, etc... to work with.
        
        Returns
        -------
        self : `None`, ``AuditLogEntry``
            Returns `None` if Discord is derping like hell.
        """
        raw_event_type = data.get('action_type', None)
        if (raw_event_type is None):
            return None
        
        self = object.__new__(cls)
        self._parent_reference = parent._get_self_reference()
        self.id = int(data['id'])
        
        event_type = AuditLogEvent.get(raw_event_type)
        self.type = event_type
        
        options = data.get('options', None)
        if (options is None):
            details = None
        else:
            details = {}
            for key, value in options.items():
                result = DETAIL_CONVERTERS.get(key, detail_converter_nothing)(key, value)
                if result is None:
                    continue
                
                key, value = result
                details[key] = value
            
            if not details:
                details = None
        
        self.details = details
        
        user_id = data.get('user_id', None)
        if user_id is None:
            user = None
        else:
            user = parent.users.get(int(user_id), None)
        self.user = user
        
        self.reason = data.get('reason', None)
        
        change_datas = data.get('changes', None)
        if (change_datas is None):
            changes = None
        else:
            changes = []
            change_converters = event_type.target_type.change_converters
            
            for change_data in change_datas:
                try:
                    key = change_data['key']
                except KeyError: # malformed?
                    continue
                
                try:
                    converter = change_converters[key]
                except KeyError:
                    if ALLOW_DEBUG_MESSAGES:
                        call_debug_logger(
                            (
                                f'Unknown Dispatch event change key: {key!r}\n'
                                f'- Change data: {data!r}\n'
                                f'- Event type: {event_type!r}'
                            ),
                            True,
                        )
                    
                    converter = MERGED_CONVERTERS.get(key, convert_nothing)
                
                change = converter(key, change_data)
                if (change is not None):
                    changes.append(change)
            
            if not changes:
                changes = None
        
        self.changes = changes
        
        target_id = data.get('target_id', None)
        if target_id is None:
            target_id = 0
        else:
            target_id = int(target_id)
        
        self.target_id = target_id
    
    
    @property
    def created_at(self):
        """
        When the audit log entry was created.
        
        Returns
        -------
        created_at : `datetime`
        """
        return id_to_datetime(self.id)
    
    
    @property
    def target(self):
        """
        Tries to resolve the target entity of the audit log entry.
        
        If the entity is not cached, returns `None`.
        
        Returns
        -------
        target : `None`, ``Guild``,  ``Channel``, ``ClientUserBase``, ``Role``, ``Webhook``,
            ``Emoji``, ``Message``, ``Integration``, ``Sticker``, ``Stage``, ``ScheduledEvent``
        """
        return self.type.target_type.target_converter(self)
    
    
    @property
    def parent(self):
        """
        Returns the entry's parent audit log.
        
        Returns
        -------
        parent : `None`, ``AuditLog``
        """
        return self._parent_reference()
    
    
    def __repr__(self):
        """Returns the representation of the audit log entry."""
        repr_parts = [
            '<', self.__class__.__name__,
            ' id=', repr(self.id),
            ', type=', self.type.name,
        ]
        
        repr_parts.append(', user=')
        user = self.user
        if user is None:
            user_repr = 'None'
        else:
            user_repr = user.full_name
        repr_parts.append(user_repr)
        
        repr_parts.append(', target=')
        repr_parts.append(repr(self.target))
        
        repr_parts.append(', change count=')
        changes = self.changes
        if changes is None:
            change_amount_repr = '0'
        else:
            change_amount_repr = repr(len(self.changes))
        repr_parts.append(change_amount_repr)
        
        reason = self.reason
        if reason is not None:
            repr_parts.append(', reason=')
            # use repr to escape special inserted characters
            repr_parts.append(repr(reason))
        
        details = self.details
        if details is not None:
            repr_parts.append(', details=')
            repr_parts.append(repr(details))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)



__all__ = ('EmbeddedActivityConfiguration',)

from scarletio import RichAttributeErrorBaseType

class EmbeddedActivityConfiguration(RichAttributeErrorBaseType):
    """
    Configuration of an embedded activity.
    
    Attributes
    ----------
    premium_tier_treatment_default : `int`
        Default value of premium tier treatment for missing keys.
    premium_tier_treatment_map : `None`, `dict` of (`int`, `int`) items
        ???
    required_premium_tier : `int`
        The required boost level of a guild to start the activity.
    """
    __slots__ = ('premium_tier_treatment_default', 'premium_tier_treatment_map', 'required_premium_tier',)
    
    def __new__(cls, data):
        """
        Creates a new embedded activity configuration from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Embedded activity configuration data.
        """
        try:
            premium_tier_treatment_all = data['activity_premium_tier_treatment_map']
        except KeyError:
            premium_tier_treatment_default = 0
            premium_tier_treatment_map = None
        else:
            premium_tier_treatment_default = premium_tier_treatment_all.pop('default', 0)
            premium_tier_treatment_map = None
            
            for key, value in premium_tier_treatment_all.items():
                if value != premium_tier_treatment_default:
                    if premium_tier_treatment_map is None:
                        premium_tier_treatment_map = {}
                    
                    premium_tier_treatment_map[int(key)] = value
        
        required_premium_tier = data.get('activity_premium_tier_level', 0)
        
        self = object.__new__(cls)
        self.premium_tier_treatment_default = premium_tier_treatment_default
        self.premium_tier_treatment_map = premium_tier_treatment_map
        self.required_premium_tier = required_premium_tier
        return self
    
    
    def __repr__(self):
        """Returns the embedded activity configuration's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        field_added = False
        
        premium_tier_treatment_default = self.premium_tier_treatment_default
        if premium_tier_treatment_default:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' premium_tier_treatment_default=')
            repr_parts.append(repr(premium_tier_treatment_default))
        
        
        premium_tier_treatment_map = self.premium_tier_treatment_map
        if (premium_tier_treatment_map is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' premium_tier_treatment_map=')
            repr_parts.append(repr(premium_tier_treatment_map))
            
        
        required_premium_tier = self.required_premium_tier
        if required_premium_tier:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' required_premium_tier=')
            repr_parts.append(repr(required_premium_tier))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the embedded activity configuration's hash value."""
        hash_value = 0
        
        premium_tier_treatment_default = self.premium_tier_treatment_default
        if premium_tier_treatment_default:
            hash_value ^= premium_tier_treatment_default << 4
        
        premium_tier_treatment_map = self.premium_tier_treatment_map
        if (premium_tier_treatment_map is not None):
            hash_value ^= hash(tuple(premium_tier_treatment_map.items()))
        
        required_premium_tier = self.required_premium_tier
        if required_premium_tier:
            hash_value ^= required_premium_tier
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two embedded activity configurations are the same."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.premium_tier_treatment_default != other.premium_tier_treatment_default:
            return False
        
        if self.premium_tier_treatment_map != other.premium_tier_treatment_map:
            return False
        
        if self.required_premium_tier != other.required_premium_tier:
            return False
        
        return True
